/* **************************************************************
 *
 *	The Album model (singleton)
 * 
 *	(c) Laszlo Molnar, 2015
 *	Licensed under Creative Commons Attribution-NonCommercial-ShareAlike 
 *	<http://creativecommons.org/licenses/by-nc-sa/3.0/>
 *
 */

 
// Constants: jAlbum conventions
var J = {
	FOLDERS: 		'folders',
	NAME: 			'name',
	PATH: 			'path',
	THUMB: 			'thumb',
	IMAGE: 			'image',
	WIDTH: 			'width',
	HEIGHT: 		'height',
	ORIGINAL: 		'original',
	FOLDERS: 		'folders',
	OBJECTS: 		'objects',
	FILEDATE: 		'fileDate',
	COMMENT: 		'comment',
	TITLE: 			'title',
	COUNTERS: 		'counters',
	DEEPCOUNTERS: 	'deepCounters',
	FILESIZE: 		'fileSize',
	CATEGORY: 		'category',
	KEYWORDS: 		'keywords',
	CAMERA: 		'camera',
	VIDEO: 			'video',
	// extra vars
	LEVEL: 			'level',
	PATHREF: 		'pathRef',
	PARENTREF: 		'parentRef',
	RELPATH: 		'relPath',
	READY: 			'ready',
	SELECTED: 		'selected',
	ACTIVE:			'active',
	IMAGECAPTION: 	'imageCaption',
	THUMBCAPTION: 	'thumbCaption',
	PHOTODATA: 		'photodata',
	LOCATION: 		'location',
	SHOP:			'shop',
	MODIFIEDDAYS: 	'modifiedDays',
	MOSTPHOTOS: 	'mostphotos',
	OBJ: 			'obj'
};

// Singleton implementation

var Album = (function($) {
	
	var instance,
		settings = {
			// Name of the tree file
			treeFile: 'tree.json',
			// Name of the folder data
			dataFile: 'data1.json',
			// index file name
			indexName: 'index.html',
			// Name of the slides folder
			slidesDir: 'slides',
			// Path to root (eg. "../../")
			rootPath: '',
			// Relative path to the current folder (eg. "folder/subfolder") 
			relPath: '',
			// Retry interval and timeout
			// pollInterval: 15,
			// timeout: 5000,
			// Lazy load :: loads folder data only when requested or at the time of initialization
			lazy: true
		},
		// Texts translated
		text = getTranslations({
			and:	'and'
		}),
		
	
		// Global variables
		// The container for the whole tree
		tree = {},
		// Collection of all album paths in order to be able to store only the references
		paths = [],
		// Path to current folder
		currentFolder,
		// Currently selected item
		current,
		// Collection the JSON promises
		defer,
		// Album ready state
		ready,
		
		// Filename for images, originalname for other
		
		getItemName = function(o) {
			if (o[J.CATEGORY] === 'video') {
				var p = o[J.VIDEO][J.PATH];
				return p.substring(p.lastIndexOf('/') + 1);
			} else if (o.hasOwnProperty(J.ORIGINAL)) {
				return o[J.ORIGINAL][J.PATH];
			}
			return o[J.NAME];
		},
	
		// Getting path reference either by searching or adding as new
		
		getPathRef = function(p) {
			var idx = $.inArray(p, paths);
			if (idx >= 0) {
				return idx;
			}
			return paths.push(p) - 1;
		},
		
		// Getting a path by reference index
		
		getPath = function(idx) {
			return paths[idx];
		},
			
		// Get the pointer to a folder in a tree from a path
		
		getPointer = function(path) {
			
			if (!path) {
				// root
				return tree;
			}
			
			// Referencing by index
			if ($.isNumeric(path)) {
				if (path > paths.length) {
					console.log('Error: out of bounds path reference!');
					return null;
				}
				path = paths[path];
			}
			
			// Removing the closing slash
			if (path[path.length - 1] === '/') {
				path = path.substring(0, path.length - 1);
			}
			
			var t = tree,
				p = path.split('/'),
				i, j, k, l;
				
			for (i = 0, k = p.length; i < k; i++) {
				if (t.hasOwnProperty(J.FOLDERS)) {
					for (j = 0, l = t[J.FOLDERS].length; j < l; j++) {
						if (t[J.FOLDERS][j][J.PATH] === p[i]) {
							// found!
							t = t[J.FOLDERS][j];
							break;
						}
					}
					if (j === l) {
						// not found
						return null;
					}
				} else {
					// no subfolders
					return null;
				}
			}
			return (i === k)? t : null;
		},
		
		// Getting the parent
		
		getParent = function(o) {
			if (o === tree) {
				// At top level
				return null;
			}
			var p;
			if (o.hasOwnProperty(J.PARENTREF)) {
				// folder
				p = (o[J.PARENTREF] === null)? null : getPointer(o[J.PARENTREF]);
			} else {
				// image
				p = getPointer(o[J.PATHREF]);
			}
			// Avoid endless loops
			return (p === o)? null : p;
		},
		
		// Retrieving a single property of an element with fallback to upper level folders
		
		getInheritedProperty = function(o, a) {
			do {
				if (o.hasOwnProperty(a)) {
					return o[a];
				}
			} while (o = Album.getParent(o));
			
			return null;
		},
		
		// Retrieving a structured property (coded with JSON.stringify) 
		// of an element with fallback to upper level folders
		
		getParsedInheritedProperty = function(o, a) {
			var p = {};
			do {
				if (o.hasOwnProperty(a)) {
					p = extend(JSON.parse(o[a]), p);
				}
			} while (o = Album.getParent(o));
			
			return p;
		},
		
		// Adding level, parent pointers and relative paths for easier navigation
		
		addExtras = function() {
			
			var add = function(o, level, path) {
				
				// Level
				o[J.LEVEL] = level;
				
				if (path.length) {
					path += '/';
				}
				var op = level? o[J.PATH] : '';
				
				// path reference
				o[J.PATHREF] = getPathRef(path + op);
				
				// parent ref
				o[J.PARENTREF] = level? getPathRef(path) : null;
				
				// relative path
				var p;
				if (!settings.relPath.length) {
					// root
					p = path + op;
				} else if (o === currentFolder) {
					// current folder
					p = '';
				} else if (op.indexOf(settings.relPath) === 0) {
					// subfolder
					p = op.substring(settings.relPath.length);
				} else {
					// other branch
					p = settings.rootPath + '/' + (path + op);
				}
				o[J.RELPATH] = getPathRef(p);
				
				if (o.hasOwnProperty(J.FOLDERS)) {
					pr = getPathRef(path);
					for (var i = 0, l = o[J.FOLDERS].length; i < l; i++) {
						add(o[J.FOLDERS][i], level + 1, path + op);
					}
				}
			};
			
			add(tree, 0, '');
		},
		
		// Loading tree.json from the top level folder
		
		loadTree = function(doneFn) {
			
			//console.log('loadTree() :: ' + settings.rootPath + settings.treeFile);
					
			return $.getJSON((settings.rootPath? (settings.rootPath + '/') : '') + settings.treeFile)
				.done(function(d) {
					
					// console.log('... tree loaded'); // Tree loaded
					tree = d;
					// Getting the pointer to the current folder
					currentFolder = getPointer(settings.relPath);
					if (currentFolder === null) {
						//console.log(settings.relPath);
						console.log('Error: can\'t find folder "' + settings.relPath + '" in the database! Reverting to root folder.');
						currentFolder = '';
					}
					//console.log(currentFolder);
					// Adding extra variables
					addExtras();
					
					// Calling "done" function
					if ($.isFunction(doneFn)) {
						doneFn.call(this);
					}
				})
				.fail(function() {
					console.log('Error loading "' + settings.treeFile + '". Please note, JSON support is required for this album, but it\'s blocked in Chrome and IE when viewing local albums. The uploaded album will work fine, however.');
					
					// Calling "done" function
					if ($.isFunction(doneFn)) {
						doneFn.call(this);
					}
				});
		},
			
		// Loading a folder's detailed data
		
		loadData = function(f, doneFn) {
			
			//console.log('loadData("' + f[J.NAME] + '") + ready:' + f[J.READY]);
			
			// Couldn't identify/find a folder
			if (!f) {
				console.log('Error: loadData("null")!');
				return;
			}	
			
			// Loading the folder's objects
			if (f.hasOwnProperty(J.OBJECTS)) {
				
				// already loaded
				if ($.isFunction(doneFn)) {
					doneFn.call(this, f[J.OBJECTS]);
				}
				return true;
					
			} else {
				
				var p = getPath(f[J.RELPATH]);
			
				// we need to load it
				//console.log('Loading "'+ p + '/' + settings.dataFile + '"');
				// building defer array to be able to check the full load
				if ( !defer ) {
					defer = [];
				}
				
				defer.push( $.getJSON( (p? (p + '/') : '') + settings.dataFile ).done(function(d) {
					//console.log("data loaded for: " + f[J.NAME]);
					// If the folder has objects, let's load it
					if (d.hasOwnProperty(J.OBJECTS)) {
						f[J.OBJECTS] = d[J.OBJECTS];
						// Adding absolute and relative paths
						for (var i = 0, l = f[J.OBJECTS].length; i < l; i++) {
							f[J.OBJECTS][i][J.PATHREF] = f[J.PATHREF];
							f[J.OBJECTS][i][J.RELPATH] = f[J.RELPATH];
							if (f[J.CATEGORY] === 'folder') {
								f[J.OBJECTS][i][J.PARENTREF] = f[J.PARENTREF];
							}
						}
					}
					if ($.isFunction(doneFn)) {
						doneFn.call(this, f[J.OBJECTS]);
					}
				}).fail(function() {
					console.log('Error loading folder data: "' + (p? (p + '/') : '') + settings.dataFile + '".');
					if ($.isFunction(doneFn)) {
						doneFn.call(this, f[J.OBJECTS]);
					}
				}) );
			}
		},
		
		// Loading data for a single folder
		
		loadFolder = function(f, deep) {
			
			//console.log('loadFolder("' + f[J.NAME] + '")');
			loadData(f);
			
			if (deep && f.hasOwnProperty(J.FOLDERS)) {
				
				for (var i = 0, l = f[J.FOLDERS].length; i < l; i++) {
					loadFolder(f[J.FOLDERS][i]);
				}
			}
		},
		
		// Initializing
		
		init = function(set) {
			
			if (instance) {
				return;
			}
			
			instance = new Date();
			
			//console.log('init()');
			$.extend(settings, set);
			ready = false;
			
			if (settings.rootPath === '.') {
				settings.rootPath = '';
			}
			
			// Loading the folder's Objects: current or all
			
			var treeReady = function() {
				
				defer = [];
				
				loadFolder(settings.lazy? currentFolder : tree, !settings.lazy);
				
				//console.log(defer);
				
				// has subfolders: waiting for AJAX requests to be completed
				$.when.apply($, defer).done(function() {
					
					var d = new Date();
					console.log(defer.length + ' folder(s) loaded: ' + (d - instance) + 'ms');	
					ready = true;
					defer = null;
					current = (currentFolder && currentFolder.hasOwnProperty(J.OBJECTS))? 0 : null;
					
					if ($.isFunction(settings.ready)) {
						//console.log('calling ' + settings.ready);
						settings.ready.call(this);
					}
					
				});
							
			};
			
			// Loading the tree.json first
			
			loadTree(treeReady);
			
			return instance;
					
		},
		
		// Get images: returns only the images
		
		getImages = function() {
			
			var items = [];
		
			if (currentFolder && currentFolder.hasOwnProperty(J.OBJECTS)) {
				
				var o = currentFolder[J.OBJECTS];
				
				for (var i = 0, l = o.length; i < l; i++) {	
					if ('image.audio.video.other'.indexOf(o[i][J.CATEGORY]) >= 0) {
					//if (o[i][J.CATEGORY] === 'image' || o[i][J.CATEGORY] === 'video') {
						items.push(o[i]);
					}
				}
			}
						
			return items;
		},
		
		// Collect items by date
		
		collectByDate = function(options) {
			
			//console.log('collectByDate(' + options + ')');
			if (typeof options === UNDEF || !options.hasOwnProperty('range')) {
				return;
			}
			
			var options = $.extend({
					sort:	true,
					depth: 	'current' // 'tree' | 'current' | 'subfolders'
				}, options),
				start,
				end, 
				items = [], 
				f = (options.depth === 'tree')? tree : currentFolder,
				
				_findByDate = function(obj) {
					
					// Find images fall into the date range
					
					if (!obj || !obj.length) {
						return;
					}
					
					var i, l = obj.length,
						o, j, k;
					
					//console.log(obj);
					
					for (i = 0; i < l; i++) {
						
						o = obj[i];
						if (o[J.CATEGORY] === 'webPage' || o[J.CATEGORY] === 'folder') {
							continue;
						}
						
						if (o.hasOwnProperty(J.MODIFIEDDAYS)) {
							if (o[J.MODIFIEDDAYS] >= start && o[J.MODIFIEDDAYS] <= end) {
								items.push(o);
							}
						}
					}
				},
				
				_sort = function() {
					items.sort(function(a, b) {
						return b[J.MODIFIEDDAYS] - a[J.MODIFIEDDAYS];
					});
				},
			
				_addFolder = function(f) {
					
					// Adds one folder
					
					loadData(f, _findByDate);
					
					if (options.depth !== 'current' && f.hasOwnProperty(J.FOLDERS)) {
						// recursive to subfolders
						for (var i = 0, l = f[J.FOLDERS].length; i < l; i++) {
							_addFolder(f[J.FOLDERS][i]);
						}
					}
				};
			
			if (options.end) {
				// Absolute
				end = options.end;
				start = end - options.range;
			} else {
				// Relative
				start = Math.ceil((new Date()) / 86400000);
				start -= (options.hasOwnProperty('start'))? options.start : options.range;
				end = start + options.range;
			}
			
			// Starting a new promise collect
			
			defer = [];
			
			// Adding folder(s)
			
			_addFolder( (options.depth === 'tree')? tree : currentFolder );
			
			if ($.isFunction(options.ready)) {
				if (defer.length) {
					$.when.apply($, defer).done(function() {
						if (options.sort) {
							_sort();
						}
						options.ready.call(items, options);
					});
				} else {
					if (options.sort) {
						_sort();
					}
					options.ready.call(items, options);
				}
			}
		},
		
		// Collect search results
		
		collectItems = function(options) {
			
			//console.log('collectItems(' + set + ')');
			if (typeof options === UNDEF || !options.hasOwnProperty('terms')) {
				return;
			}
			
			var options = $.extend({
					fields: 		'creator,keywords,title,comment',
					depth: 			'current', // 'tree' | 'current' | 'subfolders'
					exact: 			false
				}, options),
				items = [], 
				fields = options.fields.split(','), 
				fieldslength = fields.length,
				terms,
				termslength,
				conjunctive = false,
				f = (options.depth === 'tree')? tree : currentFolder,
				
				_searchItems = function(obj) {
					
					// Adds fields from objects array
					
					if (!obj || !obj.length) {
						return;
					}
					
					var i, 
						l = obj.length,
						o, 
						j, 
						k;
					
					// console.log(obj);
					
					for (i = 0; i < l; i++) {
						// Iterating through the objects
						
						o = obj[i];
						if (o[J.CATEGORY] === 'webPage' || o[J.CATEGORY] === 'folder') {
							continue;
						}
						
						(function() {
							var found = conjunctive? new Array(termslength) : null;
							for (j = 0; j < fieldslength; j++) {
								if (o.hasOwnProperty(fields[j])) {
									for (k = 0; k < termslength; k++) {
										//console.log(o[fields[j]] + ' :: ' + terms[k]);
										if (terms[k]) {
											if (options.exact && terms[k].testMatch(o[fields[j]]) || terms[k].testIn(o[fields[j]])) {
												// Found a match
												if (!conjunctive) {
													// Done, adding item
													//console.log('"' + terms[k] + '" was found in: "' + o[fields[j]] + '"');
													items.push(o);
													return;
												} else {
													found[k] = true;
												}
											}
										}
									}
								}
							}
							if (conjunctive && allTrue(found)) {
								// all terms found
								items.push(o);
							}
						})();
					}
				},
			
				_addFolder = function(f) {
					
					// Adds one folder
					
					loadData(f, _searchItems);
					
					if (options.depth !== 'current' && f.hasOwnProperty(J.FOLDERS)) {
						// recursive to subfolders
						for (var i = 0, l = f[J.FOLDERS].length; i < l; i++) {
							_addFolder(f[J.FOLDERS][i]);
						}
					}
				};
			
			// Exact search with quotes: "something exact"
			if (options.terms[0] === '"' && options.terms[options.terms.length - 1] === '"') {
				terms = options.terms.substring(1, options.terms.length - 1);
				options.exact = true;
			} else {
				terms = options.terms.replace(/\s+/g, ",");
			}	
			
			// Preparing conjunctive search
			if (~terms.indexOf(',' + text.and + ',')) {
				terms = terms.replace(new RegExp(',' + text.and + ',', 'gi'), ',');
				conjunctive = true;
			}
				
			terms = terms.split(',');
			termslength = terms.length;
			
			// Starting a new promise collect
			
			defer = [];
			
			// Adding folder(s)
			
			_addFolder( (options.depth === 'tree')? tree : currentFolder );
			
			if ($.isFunction(options.ready)) {
				if (defer.length) {
					$.when.apply($, defer).done(function() {
						options.ready.call(items, options);
					});
				} else {
					options.ready.call(items, options);
				}
			}
		},
		
		// Tag cloud
		
		collectTags = function(options) {
			
			//console.log('collectTags(' + set + ')');
				
			var options = $.extend({
				fields: 'creator,keywords',
				types:	'folder,page,image',
				depth: 	'current' // 'tree' | 'current' | 'subfolders'
			}, options);
			
			var tags = [], 
				fields = $.isArray(options.fields)? options.fields : options.fields.split(","), 
				fieldslength = fields.length,
				types = $.isArray(options.types)? options.types : options.types.split(","),
				sortByName = options.sort === 'name',
				f = (options.depth === 'tree')? tree : currentFolder;
				
			var _addKey = function(key) {
				
				var k, ka = key.toString().split(',');
				
				for (var i = 0, l = ka.length; i < l; i++) {
				
					k = ka[i].toString().trim();
					if (!k.length) {
						continue;
					}
					// Adds one key or increments the counter if already added
					//console.log('addKey(' + k + '[' + (typeof k) + '])');
					
					for (var t = 0, tl = tags.length; t < tl; t++) {
						//console.log('[' + t + '] ' + tags[t][0] + '[' + typeof(tags[t][0])  + '] (' + ((tags[t][0] == k)? 'found' : 'nope') + ')');
						if (tags[t][0] === k) {
							break;
						}
					}
					if (t === tl) {
						tags.push([ k, 1 ]);
					} else {
						tags[t][1]++;
					}
				}
			};
			
			var _addFields = function(obj) {
				
				// Adds fields from objects array
				
				if (!obj || !obj.length) {
					return;
				}
				
				var i, 
					l = obj.length,
					o, 
					j;
				
				for (i = 0; i < l; i++) {
					
					o = obj[i];
					
					for (j = 0; j < fieldslength; j++) {
						if (o.hasOwnProperty(fields[j])) {
							_addKey(o[fields[j]]);
						}
					}
				}
			};
			
			var _addFolder = function(f) {
				
				// Adds one folder
				
				loadData(f, _addFields);
				
				if (options.depth !== 'current' && f.hasOwnProperty(J.FOLDERS)) {
					// recursive to subfolders
					for (var i = 0, l = f[J.FOLDERS].length; i < l; i++) {
						_addFolder(f[J.FOLDERS][i]);
					}
				}
			}
					
			var _sortTags = function() {
				
				tags.sort(function(a, b) {
					return sortByName? ('' + a[0]).localeCompare('' + b[0]) : (b[1] - a[1]);	
				});
			}
			
			// Starting a new promise collect
			
			defer = [];
			
			// Adding folder(s)
			
			_addFolder( (options.depth === 'tree')? tree : currentFolder );
			
				
			if ($.isFunction(options.ready)) {
			
				if (defer.length) {
					$.when.apply($, defer).done(function() {
						if (options.sort) {
							_sortTags();
						}
						options.ready.call(tags, options);
					});
				} else {
					if (options.sort) {
						_sortTags();
					}
					options.ready.call(tags, options);
				}
			}
					
		};
	
	//console.log("Album initialized!");
	
	return {
		
		// Initializing (loading the tree)
		init: init,
		
		// Debug functions
		
		// Album.getTree() :: Returns the whole internal tree object
		getTree: function() { 
				return tree; 
			},
		
		getMakeDate: function() {
				return new Date(tree[J.FILEDATE]);
			},
		
		getAlbumTitle: function() {
				return tree[J.TITLE] || tree[J.NAME];
			},
		
		getPaths: function() {
				return paths;
			},
		getCurrentFolder: function() { 
				return currentFolder; 
			},
		
		// Sharp
		getObjects: function() {
				return currentFolder.hasOwnProperty(J.OBJECTS)? currentFolder[J.OBJECTS] : []; 
			},
		getImages: getImages,	
		collectTags: collectTags,
		collectItems: collectItems,
		collectByDate: collectByDate,
		getFolderPath: function(o) {
				return getPath(o[J.PATHREF] || 0);
			},
		getRelativeFolderPath: function(o) {
				return getPath(o[J.RELPATH] || 0);
			},
		getParent: getParent, 
		getItemName: getItemName,
		getItemPath: function(o) {
				var p = getPath(o[J.RELPATH]),
					c = o[J.CATEGORY] || 'image';
				p = p.length? (p + '/') : '';	
				if (c === 'video') {
					return p + o[J.VIDEO][J.PATH];
				} else if (c === 'audio' || c === 'other' || o.hasOwnProperty(J.ORIGINAL)) {
					return p + o[J.ORIGINAL][J.PATH];
				}
				return p + o[J.IMAGE][J.PATH];
			},
		getThumbPath: function(o) {
				var p = getPath(o[J.RELPATH]);
				return (p.length? (p + '/') : '') + o[J.THUMB][J.PATH];
			},
		getImagePath: function(o) {
				var p = getPath(o[J.RELPATH]);
				p = p.length? (p + '/') : '';
				return p + o[J.IMAGE][J.PATH];
			},
		getSourcePath: function(o) {
				var p = getPath(o[J.RELPATH]);
				return (p.length? (p + '/') : '') + (o.hasOwnProperty(J.ORIGINAL)? o[J.ORIGINAL][J.PATH] : o[J.IMAGE][J.PATH]);
			},
		getAbsolutePath: function(o) {
				var p = getPath(o[J.RELPATH]);
				return (p.length? p.fullUrl() : window.location.href.getDir()) + (window.location.href.getFile() || settings.indexName)  + '#img=' + getItemName(o);
			},
		getAbsoluteImagePath: function(o) {
				var p = getPath(o[J.RELPATH]);
				return (p.length? p.fullUrl() : window.location.href.getDir()) + o[J.IMAGE][J.PATH];
			},
		isImage: function(o) {
				return o.hasOwnProperty(J.CATEGORY) && o[J.CATEGORY] === 'image';
			},
		isAudio: function(o) {
				return o.hasOwnProperty(J.CATEGORY) && o[J.CATEGORY] === 'audio';
			},
		isVideo: function(o) {
				return o.hasOwnProperty(J.CATEGORY) && o[J.CATEGORY] === 'video';
			},
		isCurrentFolder: function(o) {
				return o[J.RELPATH] === 0;
			},
		getTitle: function(o) {	
				return o[J.TITLE] || '';
			},
		getName: function(o) {
				return o[J.NAME] || '';
			},
		getComment: function(o) {
				return o[J.COMMENT] || '';
			},
		getProperty: function(o, a, inherit) {
				return o.hasOwnProperty(a)? o[a] : (inherit? getInheritedProperty(o, a) : null);
			},
		getRootProperty: function(a) {
				return tree.hasOwnProperty(a)? tree[a] : null;
			},
		getParsedProperty: function(o, a, inherit) {
				return inherit? getParsedInheritedProperty(o, a) : (o.hasOwnProperty(a)? JSON.parse(o[a]) : null);
			},
		getParsedRootProperty: function(a) {
				return tree.hasOwnProperty(a)? JSON.parse(tree[a]) : null;
			}
	};
		
}(jQuery));